/**
 * Copyright (C) 2012  BonitaSoft S.A.
 * BonitaSoft, 31 rue Gustave Eiffel - 38000 Grenoble
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.util.xcmis;

import java.io.IOException;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.List;

import org.ow2.bonita.env.EnvConstants;
import org.ow2.bonita.env.Environment;
import org.ow2.bonita.env.EnvironmentFactory;
import org.ow2.bonita.env.GlobalEnvironmentFactory;
import org.ow2.bonita.facade.exception.DocumentAlreadyExistsException;
import org.ow2.bonita.facade.exception.DocumentationCreationException;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.identity.auth.DomainOwner;
import org.ow2.bonita.services.CommandService;
import org.ow2.bonita.services.Document;
import org.ow2.bonita.services.DocumentationManager;
import org.ow2.bonita.util.Command;
import org.ow2.bonita.util.jdbc.SqlResult;
import org.ow2.bonita.util.jdbc.SqlSession;
import org.ow2.bonita.util.jdbc.SqlSessionException;
import org.ow2.bonita.util.jdbc.SqlSessionManager;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * @author Elias Ricken de Medeiros
 *
 */
public class XCMISReader {

  private static final String PROCESS_DEFINITION_PATTERN = "(.+)(--)(.+)";

  private static final String PROCESS_INSTANCE_PATTERN = "(.*)(--)(.*)--(\\d+)";

  private static final Logger LOG = LoggerFactory.getLogger(XCMISReader.class);

  private final Configuration configuration;

  private final String dbUrl;

  private DocumentationManager documentationManger;
  private final String dbDriver;
  private final String dbUsername;
  private final String dbPassword;

  private final String repositoryName;

  private EnvironmentFactory envFactory;
  
  private final List<DocumentLink> documentLinks;

  private final String domain;

  public XCMISReader(final String dbUrl, final String dbDriver, final String dbUsername, final String dbPassword, final String repositoryName, final String domain) throws ConfigurationException, IOException {
    this.dbUrl = dbUrl;
    this.dbDriver = dbDriver;
    this.dbUsername = dbUsername;
    this.dbPassword = dbPassword;
    this.repositoryName = repositoryName;
    this.domain = domain;
    this.configuration = new Configuration(dbDriver);
    this.documentLinks = new ArrayList<DocumentLink>(5);
  }

  public void readDataBase() throws Exception {
    final SqlSession sqlSession = getSession();
    if (sqlSession != null) {
      envFactory = GlobalEnvironmentFactory.getEnvironmentFactory(domain);
      final Environment environment = envFactory.openEnvironment();
      try {
        DomainOwner.setDomain(domain);
        documentationManger = (DocumentationManager) envFactory.get(EnvConstants.DOCUMENTATION_MANAGER_DEFAULT_KEY);
        final String repositoryId = getRepositoryId(sqlSession, repositoryName);
        // remove [] to get the repository name
        if (repositoryId != null) {
          handleRootFolder(sqlSession, repositoryId);
          if (!documentLinks.isEmpty()) {
            handleDocumentLinks(sqlSession);
          }
        } else {
          if (LOG.isErrorEnabled()) {
            LOG.error("Unable to find repository with name " + repositoryName);
          }
        }

      } catch (final DocumentationCreationException e) {
        if (LOG.isErrorEnabled()) {
          LOG.error(e.getMessage(), e);
        }
      } finally {
        try {
          sqlSession.disconnect();
        } catch (final SqlSessionException e) {
          if (LOG.isErrorEnabled()) {
            LOG.error("Error while disconnecting from dabase", e);
          }
        }
        DomainOwner.setDomain(null);
        environment.close();
      }
    }

  }

  private SqlSession getSession() {
    SqlSession sqlSession = null;
    try {
      sqlSession = SqlSessionManager.getSession(dbUrl, dbDriver, dbUsername, dbPassword);
    } catch (final SqlSessionException e) {
      if (LOG.isErrorEnabled()) {
        LOG.error("Unable to connect to the dabase", e);
      }
    }
    return sqlSession;
  }

  private void handleRootFolder(final SqlSession sqlSession, final String folderId) throws DocumentAlreadyExistsException, DocumentationCreationException {
    // final long t1 = System.currentTimeMillis();
    List<String> subFolderIds = Collections.emptyList();
    long index = 0;
    try {
      do {
        // process definition
        subFolderIds = getSubFolderIds(sqlSession, folderId, index);
        // this.docTimes.add((System.currentTimeMillis() - t1));
        for (final String subForderId : subFolderIds) {
          handleProcessDefinitionFolder(sqlSession, subForderId);
        }
        index += this.configuration.getPageSize();
      } while (!subFolderIds.isEmpty());

    } catch (final Exception e) {
      throw new DocumentationCreationException("Unable to get name of folder with id " + folderId, e);
    }

  }

  private void handleProcessDefinitionFolder(final SqlSession sqlSession, final String folderId) throws DocumentAlreadyExistsException, DocumentationCreationException {
    String folderName;
    try {
      // discard two first characters to remove []
      folderName = getPropertyNameFromId(sqlSession, folderId).substring(2);
    } catch (final Exception e) {
      throw new DocumentationCreationException("Unable to get name of folder with id " + folderId, e);
    }
    if (folderName.matches(PROCESS_DEFINITION_PATTERN)) {
      final ProcessDefinitionUUID processUUID = new ProcessDefinitionUUID(folderName);
      handleDocuments(sqlSession, folderId, processUUID, null, false);

      // final long t1 = System.currentTimeMillis();
      List<String> subFolderIds = Collections.emptyList();
      long index = 0;
      try {
        do {
          subFolderIds = getSubFolderIds(sqlSession, folderId, index);
          // this.docTimes.add((System.currentTimeMillis() - t1));
          for (final String subForderId : subFolderIds) {
            handleProcessInstanceAndTemplateFolder(sqlSession, subForderId, processUUID);
          }
          index += this.configuration.getPageSize();
        } while (!subFolderIds.isEmpty());

      } catch (final Exception e) {
        throw new DocumentationCreationException("Unable to get name of folder with id " + folderId, e);
      }
    } else {
      if (LOG.isWarnEnabled()) {
        LOG.warn("-----------> The following folder does not seem to refer to a process definition and will be ignored: " + folderName);
      }
    }

  }
  
  private void handleProcessInstanceAndTemplateFolder(final SqlSession sqlSession, final String folderId, final ProcessDefinitionUUID processUUID)
      throws DocumentAlreadyExistsException, DocumentationCreationException {
    String folderName;
    try {
      folderName = getPropertyNameFromId(sqlSession, folderId).substring(2);
    } catch (final Exception e) {
      throw new DocumentationCreationException("Unable to get name of folder with id " + folderId, e);
    }
    ProcessInstanceUUID instanceUUID = null;
    boolean metaDocument = false;
    if (folderName.matches(PROCESS_INSTANCE_PATTERN)) {
      instanceUUID = new ProcessInstanceUUID(folderName);
    } else if ("template".equals(folderName)) {
      // it's a meta document
      metaDocument = true;
    } else {
      if (LOG.isWarnEnabled()) {
        LOG.warn("-----------> The following folder does not seem to refer to a process instance and will be ignored: " + folderName + " ****");
      }
    }
    handleDocuments(sqlSession, folderId, processUUID, instanceUUID, metaDocument);

  }

  private void handleDocuments(final SqlSession sqlSession, final String folderId, final ProcessDefinitionUUID processUUID, final ProcessInstanceUUID instanceUUID,
      final boolean metaDocument) throws DocumentAlreadyExistsException, DocumentationCreationException {
    // final long t1 = System.currentTimeMillis();
    if(LOG.isInfoEnabled()) {
      if (instanceUUID != null) {
        LOG.info("Migrating documents of process instance " + instanceUUID + " ...");
      } else {
        LOG.info("Migrating documents of process definition " + processUUID + " ...");
      }
    }
    List<String> documentIds = Collections.emptyList();
    long index = 0;
    do {
      documentIds = getDocumentIds(sqlSession, folderId, index);

      // this.docTimes.add((System.currentTimeMillis() - t1));
      for (final String documentId : documentIds) {
        handleDocument(sqlSession, documentId, processUUID, instanceUUID, metaDocument);
      }
      index += this.configuration.getPageSize();
    } while (!documentIds.isEmpty());

    if(LOG.isInfoEnabled()) {
      if (instanceUUID != null) {
        LOG.info("Documents of process instance " + instanceUUID + " were migrated.");
      } else {
        LOG.info("Documents of process definition " + processUUID + " were migrated.");
      }
    }
  }
  
  private void handleDocumentLinks(final SqlSession sqlSession) throws DocumentationCreationException {
    for (final DocumentLink link : documentLinks) {
      handleDocumentLink(sqlSession, link);
      
    }
  }

  private void handleDocumentLink(final SqlSession sqlSession, final DocumentLink link) throws DocumentationCreationException {
    try {
      final String actualDocumentId = getContainerName(sqlSession, link.getDocumentId()) + getPropertyStringValue(sqlSession, link.getFileContentPropertyId());
      final String documentName = getPropertyNameFromId(sqlSession, actualDocumentId).substring(2);
      final String parentId = getParentId(sqlSession, actualDocumentId);
      final String parentName = getPropertyNameFromId(sqlSession, parentId).substring(2);
      ProcessDefinitionUUID srcProcessUUID = null;
      ProcessInstanceUUID srcInstanceUUID = null;

      if (parentName.matches(PROCESS_INSTANCE_PATTERN)) {
        srcInstanceUUID = new ProcessInstanceUUID(parentName);
      } else if (parentName.matches(PROCESS_DEFINITION_PATTERN)) {
        srcProcessUUID = new ProcessDefinitionUUID(parentName);
      }

      final AttachDocumentCommand command = new AttachDocumentCommand(documentName, link.getInstanceUUID(), link.getProcessUUID(), srcInstanceUUID, srcProcessUUID);

      envFactory.get(CommandService.class).execute(command);

    } catch (final SQLException e) {
      throw new DocumentationCreationException("unable to retrieve all document properties.", e);
    } catch (final SqlSessionException e) {
      throw new DocumentationCreationException("unable to retrieve all document properties.", e);
    }

  }

  private void handleDocument(final SqlSession sqlSession, final String documentId, final ProcessDefinitionUUID processUUID, final ProcessInstanceUUID instanceUUID,
      final boolean metaDocument) throws DocumentationCreationException {
    try {
      final String documentName = getPropertyNameFromId(sqlSession, documentId).substring(2);

      final String fileContentPropertyId = getPropertyIdWithName(sqlSession, documentId, CMISConstants.CONTENT_KEY);
      if (isDocumentLink(sqlSession, documentId)) {
        // if it's a link store in memory to handle latter
        documentLinks.add(new DocumentLink(documentId, processUUID, instanceUUID, fileContentPropertyId));
      } else {
        // handle versions
        final String fileVersionHistoryPropertyId = getPropertyIdWithName(sqlSession, documentId, CMISConstants.VERSION_HISTORY_KEY);
        final String fileVersionHistoryValue = "system" + getPropertyStringValue(sqlSession, fileVersionHistoryPropertyId);
        // fileVersionHistoryValue represents a propertyId

        final String fileRootVersionPropertyId = getPropertyIdWithName(sqlSession, fileVersionHistoryValue, CMISConstants.ROOT_VERSION_KEY);
        final String fileRootVersionSuccessorPropertyId = getPropertyIdWithName(sqlSession, fileRootVersionPropertyId, CMISConstants.SUCCESSORS_KEY);

        // handle successors
        String currentVersionId = fileRootVersionSuccessorPropertyId;

        Document document = null;
        while (currentVersionId != null) {
          final String currentVersionIdValue = "system" + getPropertyStringValue(sqlSession, currentVersionId);
          document = handleDocumentVersion(sqlSession, currentVersionIdValue, documentName, document, processUUID, instanceUUID, metaDocument);
          // check if there is a successor to the current version
          currentVersionId = getPropertyIdWithName(sqlSession, currentVersionIdValue, CMISConstants.SUCCESSORS_KEY);
        }

        final String fileNameValue = getPropertyStrValue(sqlSession, documentId, CMISConstants.CONTENT_STREAM_FILE_NAME_KEY);
        final String createdByValue = getPropertyStrValue(sqlSession, documentId, CMISConstants.CREATED_BY_KEY);
        final String creationDateValue = getPropertyStrValue(sqlSession, documentId, CMISConstants.CREATION_DATE_KEY);
        final String lastModificationDateValue = getPropertyStrValue(sqlSession, documentId, CMISConstants.LAST_MODIFICATION_DATE_KEY);
        final String isMajorVersionValue = getPropertyStrValue(sqlSession, documentId, CMISConstants.IS_MAJOR_VERSION_KEY);
        final String contentMimeTypeValue = getPropertyStrValue(sqlSession, documentId, CMISConstants.CONTENT_MIME_TYPE_KEY);
        final byte[] fileContentValue = getPropertyByteValue(sqlSession, fileContentPropertyId, CMISConstants.CONTENT_DATA_KEY);

        final Date creationDate = getDate(creationDateValue, true);
        final Date lastModificationDate = getDate(lastModificationDateValue, false);
        createDocumentOrVersion(documentName, document, fileNameValue, isMajorVersionValue, contentMimeTypeValue, fileContentValue, instanceUUID, processUUID,
            createdByValue, creationDate, lastModificationDate, metaDocument);
      }
    } catch (final SQLException e) {
      throw new DocumentationCreationException("unable to retrieve all document properties.", e);
    } catch (final SqlSessionException e) {
      throw new DocumentationCreationException("unable to retrieve all document properties.", e);
    }

  }

  private Document createDocumentOrVersion(final String documentName, final Document predecessor, final String fileNameValue, final String isMajorVersionValue,
      final String contentMimeTypeValue, final byte[] fileContentValue, final ProcessInstanceUUID instanceUUID, final ProcessDefinitionUUID definitionUUID,
      final String author, final Date creationDate, final Date lastModificationDate, final boolean metaDocument) throws DocumentationCreationException,
      DocumentAlreadyExistsException {
    final CreateDocumentOrVersionCommand command = new CreateDocumentOrVersionCommand(documentName, predecessor, fileNameValue, isMajorVersionValue,
        contentMimeTypeValue, fileContentValue, instanceUUID, definitionUUID, author, creationDate, lastModificationDate, metaDocument);
    final Document doc = envFactory.get(CommandService.class).execute(command);
    return doc;
  }

  private String getPropertyStrValue(final SqlSession sqlSession, final String parentId, final String propertyName) throws SqlSessionException, SQLException {
    final String propertyId = getPropertyIdWithName(sqlSession, parentId, propertyName);
    final String propertyValue = getPropertyStringValue(sqlSession, propertyId);
    return propertyValue;
  }

  private byte[] getPropertyByteValue(final SqlSession sqlSession, final String parentId, final String propertyName) throws SqlSessionException, SQLException {
    final String propertyId = getPropertyIdWithName(sqlSession, parentId, propertyName);
    final byte[] propertyValue = getPropertyByteValue(sqlSession, propertyId);
    return propertyValue;
  }

  private Document handleDocumentVersion(final SqlSession sqlSession, final String documentVersionId, final String documentName, final Document predecessor,
      final ProcessDefinitionUUID processUUID, final ProcessInstanceUUID instanceUUID, final boolean metaDocument) throws DocumentationCreationException {
    try {
      final String frozenNodePropertyId = getPropertyIdWithName(sqlSession, documentVersionId, CMISConstants.FROZEN_NODE_KEY);

      final String fileNameValue = getPropertyStrValue(sqlSession, frozenNodePropertyId, CMISConstants.CONTENT_STREAM_FILE_NAME_KEY);
      final String fileContentPropertyId = getPropertyIdWithName(sqlSession, frozenNodePropertyId, CMISConstants.CONTENT_KEY);
      final String createdByValue = getPropertyStrValue(sqlSession, frozenNodePropertyId, CMISConstants.CREATED_BY_KEY);
      final String creationDateValue = getPropertyStrValue(sqlSession, frozenNodePropertyId, CMISConstants.CREATION_DATE_KEY);
      final String lastModificationDateValue = getPropertyStrValue(sqlSession, frozenNodePropertyId, CMISConstants.LAST_MODIFICATION_DATE_KEY);
      final String isMajorVersionValue = getPropertyStrValue(sqlSession, frozenNodePropertyId, CMISConstants.IS_MAJOR_VERSION_KEY);
      // final String versionLabelValue = getPropertyStrValue(sqlSession,
      // frozenNodePropertyId, CMISConstants.VERSION_LABEL_KEY);
      final String contentMimeTypeValue = getPropertyStrValue(sqlSession, frozenNodePropertyId, CMISConstants.CONTENT_MIME_TYPE_KEY);

      // data
      final byte[] fileContentValue = getPropertyByteValue(sqlSession, fileContentPropertyId, CMISConstants.CONTENT_DATA_KEY);

      final Date creationDate = getDate(creationDateValue, true);
      final Date lastModificationDate = getDate(lastModificationDateValue, false);
      final Document current = createDocumentOrVersion(documentName, predecessor, fileNameValue, isMajorVersionValue, contentMimeTypeValue, fileContentValue,
          instanceUUID, processUUID, createdByValue, creationDate, lastModificationDate, metaDocument);

      return current;
    } catch (final SQLException e) {
      throw new DocumentationCreationException("unable to retrieve all document properties.", e);
    } catch (final SqlSessionException e) {
      throw new DocumentationCreationException("unable to retrieve all document properties.", e);
    }

  }

  private Date getDate(final String strDate, final boolean isCreationDate) {
    // remove characters from --
    if (strDate == null) {
      return new Date();
    }
    int indexOf = strDate.indexOf("--");
    String gmtDate;
    if (indexOf > 0) {
      gmtDate = strDate.substring(0, indexOf);
    } else {
      gmtDate = strDate;
    }
    indexOf = gmtDate.lastIndexOf(':');
    if (indexOf > 20) {
      gmtDate = gmtDate.substring(0, indexOf) + gmtDate.substring(indexOf + 1, gmtDate.length());
    }
    final SimpleDateFormat simpleDateFormat = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.SSSZ");
    try {
      return simpleDateFormat.parse(gmtDate);
    } catch (final ParseException e) {
      if (LOG.isWarnEnabled()) {
        final StringBuilder stb = new StringBuilder();
        stb.append("Enable to parse date from string '");
        stb.append(gmtDate);
        if (isCreationDate) {
          stb.append("'. The creation date will be set to now.");
        } else {
          stb.append("'. The last modification date will be set to now.");
        }
        LOG.warn(stb.toString());
      }
      return new Date();
    }
  }

  private String getPropertyNameFromId(final SqlSession sqlSession, final String propertyId) throws DocumentationCreationException {
    SqlResult sqlResult = null;
    try {
      sqlResult = sqlSession.select(this.configuration.getQuery("getPropertyNameFromId"), new Object[] { propertyId });
      final ResultSet resultSet = sqlResult.getResultSet();
      String name = null;
      if (resultSet.next()) {
        name = resultSet.getString("name");
      }
      return name;
    } catch (final Exception e) {
      throw new DocumentationCreationException("Unable to get name of property with id " + propertyId, e);
    } finally {
      try {
        closeSqlResult(sqlResult);
      } catch (final SqlSessionException e) {
        throw new DocumentationCreationException("Error while closing sql result", e);
      }
    }
  }

  private String getPropertyIdWithName(final SqlSession sqlSession, final String parentId, final String propertyName) throws SqlSessionException, SQLException {
    SqlResult sqlResult = null;
    try {
      sqlResult = sqlSession.select(this.configuration.getQuery("getPropertyIdWithName"), new Object[] { parentId, propertyName });
      final ResultSet resultSet = sqlResult.getResultSet();
      if (!resultSet.next()) {
        return null;
      }
      return resultSet.getString("id");
    } finally {
      closeSqlResult(sqlResult);
    }
  }

  private boolean isDocumentLink (final SqlSession sqlSession, final String parentId) throws SqlSessionException, SQLException {
    SqlResult sqlResult = null;
    try {
      sqlResult = sqlSession.select(this.configuration.getQuery("getDocumentLink"), new Object[] { parentId});
      final ResultSet resultSet = sqlResult.getResultSet();
      if (resultSet.next()) {
        return true;
      }
      return false;
    } finally {
      closeSqlResult(sqlResult);
    }
  }

  private String getContainerName (final SqlSession sqlSession, final String id) throws SqlSessionException, SQLException {
    SqlResult sqlResult = null;
    try {
      sqlResult = sqlSession.select(this.configuration.getQuery("getContainerName"), new Object[] { id});
      final ResultSet resultSet = sqlResult.getResultSet();
      String containerName = null;
      if (resultSet.next()) {
        containerName = resultSet.getString("container_name");
      }
      return containerName;
    } finally {
      closeSqlResult(sqlResult);
    }
  }

  private String getParentId (final SqlSession sqlSession, final String id) throws SqlSessionException, SQLException {
    SqlResult sqlResult = null;
    try {
      sqlResult = sqlSession.select(this.configuration.getQuery("getParentIdFromId"), new Object[] { id});
      final ResultSet resultSet = sqlResult.getResultSet();
      String containerName = null;
      if (resultSet.next()) {
        containerName = resultSet.getString("parent_id");
      }
      return containerName;
    } finally {
      closeSqlResult(sqlResult);
    }
  }

  private String getPropertyStringValue(final SqlSession sqlSession, final String propertyId) throws SqlSessionException, SQLException {
    SqlResult sqlResult = null;
    try {
      sqlResult = sqlSession.select(this.configuration.getQuery("getPropertyValue"), new Object[] { propertyId });
      final ResultSet resultSet = sqlResult.getResultSet();
      if (!resultSet.next()) {
        return null;
      }
      return new String(resultSet.getBytes("data"));
    } finally {
      closeSqlResult(sqlResult);
    }
  }

  private byte[] getPropertyByteValue(final SqlSession sqlSession, final String propertyId) throws SqlSessionException, SQLException {
    SqlResult sqlResult = null;
    try {
      sqlResult = sqlSession.select(this.configuration.getQuery("getPropertyValue"), new Object[] { propertyId });
      final ResultSet resultSet = sqlResult.getResultSet();
      if (!resultSet.next()) {
        return null;
      }
      return resultSet.getBytes("data");
    } finally {
      closeSqlResult(sqlResult);
    }
  }

  private String getRepositoryId(final SqlSession sqlSession, final String repositoryName) throws DocumentationCreationException {
    SqlResult sqlResult = null;
    final String completeName = "[]" + repositoryName; // in the database the
                                                       // repository name is
                                                       // prefixed by []
    try {
      sqlResult = sqlSession.select(this.configuration.getQuery("getRepositoryId"), new Object[] {1, completeName, CMISConstants.DRIVES_KEY, CMISConstants.SYSTEM });
      final ResultSet resultSet = sqlResult.getResultSet();

      String repositoryId = null;
      if (resultSet.next()) {
        repositoryId = resultSet.getString("id");
      }
      return repositoryId;
    } catch (final Exception e) {
      throw new DocumentationCreationException("Unable to get document repositories", e);
    } finally {
      try {
        closeSqlResult(sqlResult);
      } catch (final SqlSessionException e) {
        throw new DocumentationCreationException("Error while closing sql result", e);
      }
    }
  }

  private List<String> getSubFolderIds(final SqlSession sqlSession, final String parentFolderId, final long fromIndex) throws SqlSessionException, SQLException {
    // final long start = System.currentTimeMillis();
    SqlResult sqlResult = null;
    try {
      final Object[] queryParameters = getQueryParameters(parentFolderId, fromIndex);
      sqlResult = sqlSession.select(this.configuration.getQuery("getSubFolderIds"), queryParameters);
      final ResultSet resultSet = sqlResult.getResultSet();
      final List<String> subFolderIds = new ArrayList<String>();
      while (resultSet.next()) {
        final String id = resultSet.getString("id");
        subFolderIds.add(id);
      }
      // final long end = System.currentTimeMillis();
      // printMessage("Time to get subfolders: " + (end - start));
      return subFolderIds;
    } finally {
      closeSqlResult(sqlResult);
    }

  }
  
  private Object[] getQueryParameters(final String parentId, final long fromIndex) {
    Object[] queryParameters = null;
    if (Configuration.JDBC_ORACLE_DRIVER.equals(dbDriver)) { 
      //in the syntax used by Oracle and Postgresql it's necessary to put the page size before from index 
      queryParameters = new Object[] { parentId, fromIndex + configuration.getPageSize(), fromIndex};
    } else if ( Configuration.JDBC_POSTGRESQL_DRIVER.equals(dbDriver)) {
      //in the syntax used by Oracle and Postgresql it's necessary to put the page size before from index 
      queryParameters = new Object[] { parentId, configuration.getPageSize(), fromIndex};
    } else if (Configuration.JDBC_SQLSERVER_DRIVER.equals(dbDriver)) {
      //the query for sql server is fromIndex, fromIndex + pagesize
      queryParameters = new Object[] { parentId, fromIndex, fromIndex, configuration.getPageSize()};
    } else {
      queryParameters = new Object[] { parentId, fromIndex, configuration.getPageSize() };
    }
    return queryParameters;
  }
  
  private List<String> getDocumentIds(final SqlSession sqlSession, final String parentFolderId, final long fromIndex) throws DocumentationCreationException {
    // final long start = System.currentTimeMillis();
    SqlResult sqlResult = null;
    try {
      final Object[] queryParameters = getQueryParameters(parentFolderId, fromIndex);
      sqlResult = sqlSession.select(this.configuration.getQuery("getDocumentIds"), queryParameters);
      final ResultSet resultSet = sqlResult.getResultSet();
      final List<String> documentIds = new ArrayList<String>();
      while (resultSet.next()) {
        final String id = resultSet.getString("id");
        documentIds.add(id);
      }
      // final long end = System.currentTimeMillis();
      // System.out.println(("############### Time to get document ids: " + (end
      // - start)));
      return documentIds;
    } catch (final SQLException e) {
      throw new DocumentationCreationException("Unable to get documents ids.", e);
    } catch (final SqlSessionException e) {
      throw new DocumentationCreationException("Unable to get documents ids", e);
    } finally {
      try {
        closeSqlResult(sqlResult);
      } catch (final SqlSessionException e) {
        throw new DocumentationCreationException("Error while closing sql result", e);
      }
    }
  }

  private boolean closeSqlResult(final SqlResult sqlResult) throws SqlSessionException {
    if (sqlResult != null) {
      sqlResult.close();
      return true;
    }
    return false;
  }

  class CreateDocumentOrVersionCommand implements Command<Document> {

    private static final long serialVersionUID = -6525524729382725216L;
    private final String documentName;
    private final Document predecessor;
    private final String fileNameValue;
    private final String isMajorVersionValue;
    private final String contentMimeTypeValue;
    private final byte[] fileContentValue;
    private final ProcessInstanceUUID instanceUUID;
    private final ProcessDefinitionUUID definitionUUID;
    private final String author;
    private final Date creationDate;
    private final Date lastModificationDate;
    private final boolean metaDocument;

    public CreateDocumentOrVersionCommand(final String documentName, final Document predecessor, final String fileNameValue, final String isMajorVersionValue,
        final String contentMimeTypeValue, final byte[] fileContentValue, final ProcessInstanceUUID instanceUUID, final ProcessDefinitionUUID definitionUUID,
        final String author, final Date creationDate, final Date lastModificationDate, final boolean metaDocument) {
      this.documentName = documentName;
      this.predecessor = predecessor;
      this.fileNameValue = fileNameValue;
      this.isMajorVersionValue = isMajorVersionValue;
      this.contentMimeTypeValue = contentMimeTypeValue;
      this.fileContentValue = fileContentValue;
      this.instanceUUID = instanceUUID;
      this.definitionUUID = definitionUUID;
      this.author = author;
      this.creationDate = creationDate;
      this.lastModificationDate = lastModificationDate;
      this.metaDocument = metaDocument;
    }

    @Override
    public Document execute(final Environment environment) throws Exception {
      Document current = null;
      if (predecessor == null) {
        if (metaDocument) {
          current = documentationManger.createMetaDocument(definitionUUID, documentName, fileNameValue, contentMimeTypeValue, fileContentValue, author, creationDate,
              lastModificationDate);
        } else {
          current = documentationManger.createDocument(documentName, definitionUUID, instanceUUID, fileNameValue, contentMimeTypeValue, fileContentValue, author,
              creationDate, lastModificationDate);
        }
      } else {
        current = documentationManger.createVersion(predecessor.getId(), Boolean.valueOf(isMajorVersionValue), fileNameValue, contentMimeTypeValue, fileContentValue,
            author, creationDate, lastModificationDate);
      }
      return current;
    }

  }

  class AttachDocumentCommand implements Command<Void> {

    private static final long serialVersionUID = -2646895614868297438L;
    private final ProcessInstanceUUID instanceUUIDToAttach;
    private final ProcessDefinitionUUID definitionUUIDToAttach;
    private final String documentName;
    private final ProcessInstanceUUID srcInstanceUUID;
    private final ProcessDefinitionUUID srcDefinitionUUID;

    public AttachDocumentCommand(final String documentName, final ProcessInstanceUUID instanceUUIDToAttach, final ProcessDefinitionUUID definitionUUIDToAttach,
        final ProcessInstanceUUID srcInstanceUUID, final ProcessDefinitionUUID srcDefinitionUUID) {
      this.documentName = documentName;
      this.instanceUUIDToAttach = instanceUUIDToAttach;
      this.definitionUUIDToAttach = definitionUUIDToAttach;
      this.srcInstanceUUID = srcInstanceUUID;
      this.srcDefinitionUUID = srcDefinitionUUID;
    }

    @Override
    public Void execute(final Environment environment) throws Exception {
      String documentId = null;
      if (srcInstanceUUID == null) {
        final List<Document> documents = documentationManger.getDocuments(srcDefinitionUUID, documentName, 0, 1);
        if (!documents.isEmpty()) {
          documentId = documents.get(0).getId();
        }
      } else {
        final List<Document> documents = documentationManger.getDocuments(srcInstanceUUID, documentName, 0, 1);
        if (!documents.isEmpty()) {
          documentId = documents.get(0).getId();
        }
      }

      if (documentId != null) {
        documentationManger.attachDocumentTo(definitionUUIDToAttach, instanceUUIDToAttach, documentId);
      } else {
        if (LOG.isErrorEnabled()) {
          final StringBuilder stb = new StringBuilder();
          stb.append("Unable to migrate document: ");
          stb.append(documentName);
          if (definitionUUIDToAttach != null) {
          stb.append(", ProcessDefinitionUUID: ");
          stb.append(definitionUUIDToAttach);
          }
          if (instanceUUIDToAttach != null) {
          stb.append(", ProcessInstanceUUID: ");
          stb.append(instanceUUIDToAttach);
          }
          LOG.error(stb.toString());
        }
      }
      return null;
    }

  }
  
  
}